# Project page needs:
#  1. Basic project info (name etc.)
#  2. Currently deployed revision
#  3. Undeployed revisions, of which for each
#     a. the basic info
#     b. its status (involves relationship to other revisions
#     c. comments for that revision
#  4. Derived info like the maximum deployable revno, blocking revision
#
# What's the best interface to get that info in the fewest queries?

import operator

from django.core.urlresolvers import reverse
from django.contrib.comments.models import Comment
from django.contrib.contenttypes.models import ContentType
from django.db.models import F

from .models import (
    DeploymentRequest,
    Revision,
    )


def project_basic_info(project):
    """Get the basic information for `project`.

    :return: a dict with the basic information. Keys are 'name', 'url'.
    """
    # Should only access columns from the row itself
    # Would be great to have a way to enforce that. Perhaps do something
    # like django-factory and extract all the TextField etc. directly
    # in to a dict
    info = dict()
    info['name'] = project.name
    info['description'] = project.description or ''
    info['url'] = reverse('project_view', args=(project.name,))
    return info


def project_deployed_revno(project):
    """Get the revno of the deployed revision of `project`.

    :return: the revno of the deployed revision, or None if the
        project is not deployed.
    """
    rev = project.deployed_revision
    if rev is None:
        return None
    return rev.revno


def project_undeployed_revisions(project, deployed_revno):
    """Get the undeployed revisions of `project`.

    :param deployed_revno: the revno of `project` that is currently deployed,
        or None if no revision is currently deployed.
    :return: a list of undeployed `Revision`s. They will be ordered in ascending
        `revno` order.
    """
    return project_revisions_between(project, deployed_revno, None)


def project_revisions_between(project, start_revno, end_revno):
    """Get the revisions of `project` between two revnos.

    The revisions returned will all satisfy start_revno < revno <= end_revno,
    and will be in revno order. They are not guaranteed to cover the entire
    range.

    Either revno can be `None` to indicate that the range is open at that end.
    """
    query = Revision.objects.filter(project__exact=project,
            parent__exact=None).extra(order_by=['revno']).select_related('committer')
    if start_revno is not None:
        query = query.filter(revno__gt=start_revno)
    if end_revno is not None:
        query = query.filter(revno__lte=end_revno)
    return list(query)


def project_deployable_revisions(undeployed_revisions, children_map):
    """Get the deployable revisions of `undeployed_revisions`.

    :param undeployed_revisions: a list of undeployed revisions, with the
        earliest first.
    :param children_map: a map from `Revision.pk` to list of child revisions.
        This will be consulted to see if there are child revisions that
        need to be considered for each revision in `undeployed_revisions`.
    :return: a list of deployable `Revision`s. They will be ordered in ascending
        `revno` order.
    """
    deployable = []
    for revision in reversed(undeployed_revisions):
        children = children_map.get(revision.pk, [])
        if revision_is_deployable(revision, children):
            deployable.insert(0, revision)
        elif revision_rolled_back_by(revision, children):
            rollbacks = revision_rolled_back_by(revision, children)
            for rollback in rollbacks[:]:
                if rollback in deployable:
                    rollbacks.remove(rollback)
            if not rollbacks:
                deployable.insert(0, revision)
            else:
                deployable = []
        else:
            deployable = []
    return deployable


def project_maximum_deployable_revision(deployable_revisions):
    """Get the maximum deployable revision from `deployable_revisions`.

    :param deployable_revisions: a list of deployable revisions, in
        ascending revno order.
    :return: the revision with the highest revno that can be deployed, or
        None if there are no deployable revisions.
    """
    if len(deployable_revisions):
        return deployable_revisions[-1]
    return None


def project_revision_blocking_deployment(undeployed_revisions,
        deployable_revisions):
    """Get the revision that is blocking deployment, if any.

    :param undeployed_revisions: a list of undeployed revisions, in ascending
        revno order.
    :param deployed_revisions: the sub-list of `undeployed_revisions` that
        is deployable, in ascending revno order.
    :return: The `Revision` that is blocking deployment, or None if there are
        no revisions blocking deployment.
    """
    if len(undeployed_revisions) > 0 and len(deployable_revisions) < len(undeployed_revisions):
        return undeployed_revisions[len(deployable_revisions)]
    return None


def project_revisions_needing_qa(undeployed_revisions):
    """Find the undeployed revisions that need QA.

    :param undeployed_revisions: a list of undeployed revisions.
    :return: a list of revisions that need QA, the revisions will be
        either members of `undeployed_revisions` or child revisions
        of these revisions.
    """
    child_revisions_needing_qa = Revision.objects.filter(
            status__exact=Revision.STATUS_UNKNOWN, parent__in=undeployed_revisions)
    revisions = list(child_revisions_needing_qa)
    def is_unknown(rev):
        return rev.status == rev.STATUS_UNKNOWN
    revisions += filter(is_unknown, undeployed_revisions)
    return revisions


def deployed_revnos_of_projects_for_revisions(revisions):
    """Get the deployed revnos for the projects of `revisions`

    Given an iterable of `Revision`s this method will return a map
    of the deployed revnos for the projects that those revisions
    correspond to.

    N.B. that the revisions passed in should have their project
    attributes pre-cached to avoid O(revisions) scaling of queries.

    :param revisions: an iterable of `Revision` objects.
    :return: a map of {project pk: deployed revno} where the key
        is the primary key of a project, and the value is the
        deployed revision number of that project. If the project
        pk is not in the dict for one of the input revisions,
        it indicates that the project isn't deployed.
    """
    projects = map(operator.attrgetter('project'), revisions)
    revnos = dict(Revision.objects.filter(project__in=projects,
        pk=F('project__deployed_revision')).values_list('project', 'revno'))
    return revnos


def revision_basic_info(revision):
    """Get the basic information for `revision`.

    :return: a dict with the basic information. Keys are 'pk', 'title',
        'description', 'branch', 'revno', 'commit_sha1', 'assignee',
        'assignee_link', 'link', 'commit_message'.
    """
    info = dict()
    info['pk'] = revision.pk
    info['title'] = revision.title
    info['description'] = revision.description
    info['branch'] = revision.branch
    info['revno'] = revision.revno
    info['commit_sha1'] = revision.commit_sha1
    info['assignee'] = revision.assignee
    info['assignee_link'] = revision.assignee_link
    info['link'] = revision.link
    info['commit_message'] = revision.commit_message
    return info


def revision_summary_status(revision, children):
    """Get the status of `revision` taking in to account any children.

    If a revision is OK, but it has a child that is BAD, then the summary
    status will be BAD, as the parent can't be deployed. This function
    will get the summary status for a revision given that revision
    and its list of children (if any.)

    The algorithm goes:
      if the revision is BAD or UNKNOWN, the summary status is the same
      if there is a BAD child then the summary status is BAD
      if there is an UNKNOWN child then the summary status is UNKNOWN
      otherwise the summary status is the same as the revision status.

    :param revision: the `Revision` to calculate the summary status for.
    :param children: the list of children of `revision`.
    :return: the summary status of `revision`.
    """
    status = None
    if children:
        if revision.status in (revision.STATUS_BAD, revision.STATUS_UNKNOWN):
            status = revision.status
        elif filter(lambda c: c.status==revision.STATUS_BAD, children):
            status = revision.STATUS_BAD
        elif filter(lambda c: c.status==revision.STATUS_UNKNOWN, children):
            status = revision.STATUS_UNKNOWN
        else:
            status = revision.status
    else:
        status = revision.status
    return status


def revision_is_deployable(revision, children):
    """Whether `revision` is deployable.

    This only takes in to account the deployability of the revision and children,
    not any preceeding revisions. This means that even if this function returns
    True the revision may not be deployable because it is blocked by an earlier
    revision.

    :param revision: the `Revision` to check.
    :param children: a list of child `Revision`s to consider if the
        status of `revision` is dependent on children.
    :return: a boolean indicating whether `revision` and all `children` are
        deployable.
    """
    if revision_summary_status(revision, children) in (revision.STATUS_OK, revision.STATUS_UNTESTABLE):
        return True
    return False


def revision_rolled_back_by(revision, children):
    """Get the rollback revisions for a revision and its children.

    :param revision: the `Revision` to check.
    :param children: a list of child `Revision`s to consider.
    :return: a list of `Revision`s that are rollbacks of
        `revision` or any of `children`.
    """
    # FIXME: we probably want to make this a zero-query function by
    # pre-loading a rollback map and passing it in
    rollbacks = []
    if revision.rolled_back_by:
        rollbacks.append(revision.rolled_back_by)
    for child in children:
        rollbacks.extend(revision_rolled_back_by(child, []))
    return rollbacks


def get_child_revisions(parents):
    """Load the child revisions for `parents`.

    :param parents: a list of `Revision` objects to get the children of.
    :return: a dict mapping {pk: [children]} where the pk is the pk of the
        `Revision` from `parents`, and the list of `Revision` objects that
        are the children of that revision. If a revision has no children
        then it will not be in the dict.
    """
    query = Revision.objects.filter(
            parent__pk__in=map(operator.attrgetter('pk'), parents),
            ).extra(order_by=['branch', 'revno']).select_related('committer')
    child_map = {}
    for child in query:
        child_map.setdefault(child.parent_id, [])
        child_map[child.parent_id].append(child)
    return child_map


def get_parent_revnos(revisions):
    """Load the parent revision numbers for `revisions`.

    :param revisions: a list of `Revision` objects to get the parents of.
    :return: a dict mapping {pk: parent revno} where the pk is the pk of the
        `Revision` from `parents`, and the revno of the parent is the value.
        If a revision has no parent (i.e.  it is top-level) then it will not
        be in the dict.
    """
    parent_map = dict(Revision.objects.filter(
            pk__in=map(operator.attrgetter('pk'), revisions)).exclude(
            parent__exact=None,
            ).values_list('pk', 'parent__revno'))
    return parent_map


def comments_for_revisions(revisions):
    """Get the comments for a set of `Revision`s.

    :param revisions: an iterable of `Revision` objects to get the comments
        for.
    :return: a dict mapping {pk: [comment]} where pk is the pk of the
        `Revision` from `revisions` that the comments are for, and the
        value is a list of `Comment` objects. If the pk of a revision
        in `revisions` is not in the map then there are no comments for it.
    """
    if not revisions:
        return dict()
    content_type_id = ContentType.objects.get_for_model(revisions[0]).pk
    comments = Comment.objects.filter(content_type__exact=content_type_id,
            object_pk__in=map(operator.attrgetter('pk'), revisions)).select_related('user')
    comment_map = dict()
    for comment in comments:
        object_pk = int(comment.object_pk)
        comment_map.setdefault(object_pk, [])
        comment_map[object_pk].append(comment)
    return comment_map


def deployment_request_summary(request):
    info = dict()
    info['project'] = request.project.name
    info['date_created'] = request.date_created.strftime("%H:%M %Y-%m-%d")
    info['deploy_revision'] = revision_basic_info(request.deploy_revision)
    info['requester'] = request.requester.username
    if request.status == DeploymentRequest.STATUS_COMPLETE:
        info['completer'] = request.completer.username
        info['date_completed'] = request.date_completed.strftime("%H:%M %Y-%m-%d")
    info['link'] = reverse('deployment_request', args=[request.project.name, request.pk])
    return info


def deployment_requests(status=DeploymentRequest.STATUS_WAITING):
    requests = DeploymentRequest.objects.filter(
            status=status,
            ).select_related('project__name', 'deploy_revision__revno',
                             'requester__username', 'completer__username',
            )
    if status == DeploymentRequest.STATUS_WAITING:
        requests = requests.order_by('-date_created')
    else:
        requests = requests.order_by('-date_completed')
        requests = requests[:100]
    return map(deployment_request_summary, requests)
