from __future__ import print_function

from email.Utils import parseaddr
import os
import re
import shutil
import tempfile
from urlparse import (
    unquote,
    urlsplit,
    urlunsplit,
    )

from bzrlib import (
    errors as bzr_errors,
    revision as _mod_revision,
    transport,
    )
from bzrlib.branch import Branch
from bzrlib.plugin import load_plugins
from bzrlib.revisionspec import RevisionSpec
from celery import shared_task
from django.conf import settings
from django.contrib.auth.models import User
from django.db.models import Max
import pygit2

from revtracker.launchpad import get_launchpad
from revtracker.models import (
    Project,
    Revision,
    )


class NoSuchRevision(Exception):

    def __init__(self, branch, revid_or_revno):
        self.branch = branch
        self.revid_or_revno = revid_or_revno

    def __str__(self):
        return '%s has no revision %s' % (self.branch, self.revid_or_revno)

    def __unicode__(self):
        return unicode(str(self))


class VCS(object):

    (VCS_BAZAAR, VCS_GIT) = range(2)


class LocalRevision(object):
    """An abstraction of a local VCS revision."""

    def __init__(self, revision, revid, revno):
        self.revision = revision
        self.revid = revid
        self.revno = revno

    def __eq__(self, other):
        return self.revid == other.revid

    def __ne__(self, other):
        return self.revid != other.revid


class LocalBranch(object):
    """An abstraction of a local VCS branch."""

    @classmethod
    def get_or_create(cls, vcs, url, project_name, git_branch=None,
                      fetched_branches=None, all_revs=False):
        if vcs == VCS.VCS_BAZAAR:
            return cls._get_or_create_branch(
                    url, project_name, fetched_branches=fetched_branches,
                    all_revs=all_revs)
        else:
            # XXX cjwatson 2016-09-22: all_revs not implemented for git, but
            # possibly not needed?
            return cls._get_or_create_git_repository(
                    url, project_name, git_branch=git_branch,
                    fetched_branches=fetched_branches)

    @classmethod
    def _get_or_create_branch(cls, url, project_name, fetched_branches=None,
                              all_revs=False):
        basedir = settings.BRANCH_CACHE_DIR
        project_dir = os.path.join(basedir, project_name)
        if os.path.exists(project_dir):
            branch_to = Branch.open(project_dir)
            update = True
            if fetched_branches is not None:
                if project_dir in fetched_branches:
                    update = False
            print('Acquiring WRITE lock for {}'.format(branch_to))
            branch_to.lock_write()
            if update:
                try:
                    branch_from = Branch.open(url)
                    print('Acquiring READ lock for {}'.format(branch_from))
                    branch_from.lock_read()
                    try:
                        branch_from_revs = branch_from._revision_history()
                        branch_from_revs_count = len(branch_from_revs)
                        branch_to_revno = branch_to.revno()
                        if branch_from_revs_count >= branch_to_revno:
                            revision_id = None
                        else:
                            if (branch_to_revno + 1000 >=
                                    branch_from_revs_count):
                                revision_id = None
                            else:
                                revision_id = branch_from_revs[
                                        branch_to_revno + 1000]
                        if revision_id is not None:
                            print("Getting up to revision %d" % (
                                    branch_to_revno + 1000,))
                        branch_to.pull(branch_from, True, revision_id)
                    finally:
                        print('Releasing lock for {}'.format(branch_from))
                        branch_from.unlock()
                except:
                    print('Releasing lock for {}'.format(branch_to))
                    branch_to.unlock()
                    raise
        else:
            branch_from = Branch.open(url)
            print('Acquiring READ lock for {}'.format(branch_from))
            branch_from.lock_read()
            try:
                if not os.path.exists(basedir):
                    os.makedirs(basedir)
                to_transport = transport.get_transport(project_dir)
                to_transport.mkdir('.')
                revision_id = None
                branch_from_revs = branch_from._revision_history()
                branch_from_revs_count = len(branch_from_revs)
                if not all_revs and branch_from_revs_count > 1000:
                    revision_id = branch_from_revs[1000]
                    print("Getting up to revision 1000")
                to_dir = branch_from.bzrdir.sprout(to_transport.base,
                                            create_tree_if_local=False,
                                            source_branch=branch_from,
                                            revision_id=revision_id)
                branch_to = to_dir.open_branch()
                print('Acquiring WRITE lock for {}'.format(branch_to))
                branch_to.lock_write()
            finally:
                print('Releasing lock for {}'.format(branch_from))
                branch_from.unlock()
        if fetched_branches is not None:
            fetched_branches[project_dir] = True
        return cls(VCS.VCS_BAZAAR, branch_to, url)

    @classmethod
    def _get_or_create_git_repository(cls, url, project_name, git_branch=None,
                                      fetched_branches=None):
        if git_branch is None:
            git_branch = 'master'
        basedir = settings.BRANCH_CACHE_DIR
        project_dir = os.path.join(basedir, project_name)
        update = True
        if 'SSH_AUTH_SOCK' in os.environ:
            credentials = pygit2.KeypairFromAgent(
                    settings.GIT_SSH_DEFAULT_USERNAME)
        else:
            credentials = pygit2.Keypair(
                    settings.GIT_SSH_DEFAULT_USERNAME,
                    settings.GIT_SSH_PUBLIC_KEY, settings.GIT_SSH_PRIVATE_KEY,
                    '')
        callbacks = pygit2.RemoteCallbacks(credentials=credentials)
        if os.path.exists(project_dir):
            repository = pygit2.Repository(project_dir)
            if fetched_branches is not None:
                if project_dir in fetched_branches:
                    update = False
            if update:
                for remote in repository.remotes:
                    remote.fetch(callbacks=callbacks)
        else:
            repository = pygit2.clone_repository(
                    url, project_dir, callbacks=callbacks)
        if update:
            # This is a bit weird, but we don't care about having a
            # non-detached HEAD locally, and it's more robust than hoping we
            # can do a fast-forward merge.
            repository.checkout('refs/remotes/origin/' + git_branch)
        return cls(VCS.VCS_GIT, repository, url)

    def __init__(self, vcs, branch, branch_url):
        self.vcs = vcs
        if self.vcs == VCS.VCS_BAZAAR and not branch.is_locked():
            raise AssertionError("Branch must be read-locked.")
        self.branch = branch
        self.branch_url = branch_url

    def __str__(self):
        return str(self.branch)

    def revno_to_revid(self, revno):
        if self.vcs == VCS.VCS_GIT:
            # This would be expensive to compute, so we take care not to
            # need to do so.
            raise NotImplementedError(
                    "revno_to_revid is not implemented for git.")
        try:
            return self.branch.get_rev_id(revno)
        except bzr_errors.NoSuchRevision:
            raise NoSuchRevision(self, revno)

    def revid_to_revno(self, revid):
        if self.vcs == VCS.VCS_BAZAAR:
            try:
                return self.branch.revision_id_to_dotted_revno(revid)[0]
            except bzr_errors.NoSuchRevision:
                raise NoSuchRevision(self, revid)
        else:
            if revid is None:
                return 0
            # This isn't very git-like, but synthesising revnos from the
            # length of the left-hand ancestry means that we can order
            # commits easily.
            try:
                commit = self.branch[revid]
            except KeyError:
                raise NoSuchRevision(self, revid)
            revno = 1
            while commit.parents:
                commit = commit.parents[0]
                revno += 1
            return revno

    def revspec_to_revision(self, revspec):
        if revspec is None:
            return None
        if self.vcs == VCS.VCS_BAZAAR:
            revid = RevisionSpec.from_string(revspec).as_revision_id(
                    self.branch)
        else:
            revid = self.branch.revparse_single(revspec).id.hex
        return self.get_revision(revid)

    def null_revision(self):
        if self.vcs == VCS.VCS_BAZAAR:
            null_revid = _mod_revision.NULL_REVISION
        else:
            null_revid = None
        return LocalRevision(None, null_revid, 0)

    def tip_revision(self):
        if self.vcs == VCS.VCS_BAZAAR:
            revno, revid = self.branch.last_revision_info()
            return self.get_revision(revid, revno=revno)
        else:
            return self.revspec_to_revision('HEAD')

    def resolve_revision(self, revid, revno):
        if revid is not None and revno is None:
            revno = self.revid_to_revno(revid)
        if revno is not None:
            if revid is None:
                revid = self.revno_to_revid(revno)
        return self.get_revision(revid, revno=revno)

    def get_revision(self, revid, revno=None):
        if self.vcs == VCS.VCS_BAZAAR:
            revision = self.branch.repository.get_revision(revid)
        else:
            revision = self.branch[revid]
        if revno is None:
            revno = self.revid_to_revno(revid)
        return LocalRevision(revision, revid, revno)

    def get_commit_message(self, revision):
        return revision.revision.message

    def get_author_email(self, revision):
        if self.vcs == VCS.VCS_BAZAAR:
            return parseaddr(revision.revision.get_apparent_authors()[0])[1]
        else:
            return revision.revision.author.email

    def get_file_text(self, revision, path):
        if self.vcs == VCS.VCS_BAZAAR:
            tree = self.branch.repository.revision_tree(revision.revid)
            file_id = tree.path2id(path)
            if file_id is None:
                return None
            return tree.get_file_text(file_id, path=path)
        else:
            tree = revision.revision.tree
            if path not in tree:
                return None
            return self.branch[tree[path].id].data

    def get_previous_revision(self, revision):
        if self.vcs == VCS.VCS_BAZAAR:
            if not len(revision.revision.parent_ids):
                return None
            else:
                previous_revid = revision.revision.parent_ids[0]
                previous_revision = self.branch.repository.get_revision(
                        previous_revid)
                return LocalRevision(
                        previous_revision, previous_revid, revision.revno - 1)
        else:
            if not revision.revision.parents:
                return None
            else:
                previous_revision = revision.revision.parents[0]
                return LocalRevision(
                        previous_revision, previous_revision.id.hex,
                        revision.revno - 1)

    def get_revision_link(self, revision):
        parsed_url = urlsplit(self.branch_url.encode("ascii"))
        components = list(parsed_url)
        if self.vcs == VCS.VCS_BAZAAR:
            if parsed_url.scheme == "lp":
                components[0] = "https"
                components[1] = "bazaar.launchpad.net"
                components[2] = "/+branch/" + components[2]
            elif parsed_url.scheme == "readonly+file":
                # Affordance for the test suite: urlparse doesn't include
                # the "readonly+file" scheme in uses_netloc, but we want to
                # preserve the extra "//".
                components[2] = "//" + components[2]
            components[2] += "/revision/%d" % revision.revno
        else:
            if parsed_url.scheme == "lp":
                components[0] = "https"
                components[1] = "git.launchpad.net"
                components[2] = "/" + components[2]
            elif parsed_url.scheme.startswith("git"):
                components[0] = "https"
                # We don't want username, password, or port.
                components[1] = parsed_url.hostname
            components[2] += "/commit?id=%s" % revision.revid
        return urlunsplit(components)

    def iter_revisions(self, start_revision=None, end_revision=None,
                       tip_only=False):
        tip_revision = self.tip_revision()
        if (start_revision is not None and
                tip_revision.revno <= start_revision.revno):
            return
        if start_revision is None:
            start_revision = self.null_revision()
        if end_revision is not None:
            current_revision = end_revision
        else:
            current_revision = tip_revision
        while (current_revision is not None and
                current_revision != start_revision):
            yield current_revision
            if tip_only:
                break
            current_revision = self.get_previous_revision(current_revision)

    def find_lefthand_merger(self, merged_revid, tip_revid):
        if self.vcs == VCS.VCS_BAZAAR:
            graph = self.branch.repository.get_graph()
            return graph.find_lefthand_merger(merged_revid, tip_revid)
        else:
            last_mainline = next_mainline = tip_revid
            for commit in self.branch.walk(
                    tip_revid, pygit2.GIT_SORT_TOPOLOGICAL):
                if commit.id.hex == next_mainline:
                    last_mainline = commit.id.hex
                    if commit.parent_ids:
                        next_mainline = commit.parent_ids[0].hex
                    else:
                        next_mainline = None
                if commit.id.hex == merged_revid:
                    return last_mainline
            return None

    def add_revision_to_db(self, lp, project, revision, email_map,
                           require_qa=True, parent_revision=None):
        get_options = {}
        create_columns = {'project': project}
        if parent_revision is None:
            create_columns['parent'] = None
            create_columns['branch'] = ''
        else:
            create_columns['parent'] = parent_revision
            create_columns['branch'] = self.branch_url
        if self.vcs == VCS.VCS_BAZAAR:
            get_options['revno'] = revision.revno
            create_columns['revno'] = revision.revno
        else:
            get_options['commit_sha1'] = revision.revid
            create_columns['commit_sha1'] = revision.revid
            create_columns['revno'] = revision.revno
        try:
            if parent_revision is None:
                revision_set = project.revision_set.filter(parent=None)
            else:
                revision_set = project.revision_set.filter(
                        branch=self.branch_url)
                revision_set = revision_set.exclude(parent=None)
            db_revision = revision_set.get(**get_options)
            created = False
        except Revision.DoesNotExist:
            db_revision = Revision(**create_columns)
            created = True
        db_revision.commit_message = self.get_commit_message(revision)
        assignee_name = get_assignee(self, revision, lp, email_map)
        if assignee_name is not None:
            db_revision.assignee = assignee_name
            db_revision.committer = get_committer(assignee_name)
        db_revision.link = self.get_revision_link(revision)
        if created:
            if require_qa:
                db_revision.status = Revision.STATUS_UNKNOWN
            else:
                db_revision.status = Revision.STATUS_OK
        db_revision.save()
        return db_revision

    def unlock(self):
        if self.vcs == VCS.VCS_BAZAAR:
            print('Releasing lock for {}'.format(self.branch))
            self.branch.unlock()


def get_assignee(branch, revision, lp, email_map):
    assignee = None
    email = branch.get_author_email(revision)
    if email is not None:
        if email in email_map:
            assignee = email_map[email]
        else:
            if "." not in email:
                assignee = None
            else:
                assignee = lp.people.getByEmail(email=email)
            email_map[email] = assignee
    if assignee is None:
        return None
    return assignee.name


def get_committer(name):
    person, created = User.objects.get_or_create(username=name)
    if created:
        person.save()
    return person


def store_revisions(project, branch, revisions):
    tempdir = tempfile.mkdtemp()
    try:
        lp = get_launchpad(tempdir)
        email_map = {}
        # Walk the revisions oldest first
        for revision, sourcedep_changes in reversed(revisions):
            parent_revision = branch.add_revision_to_db(
                    lp, project, revision, email_map,
                    require_qa=not project.no_qa_of_main_branch)
            if sourcedep_changes:
                for sourcedep_branch, revision in reversed(sourcedep_changes):
                    sourcedep_branch.add_revision_to_db(
                            lp, project, revision, email_map,
                            parent_revision=parent_revision)
    finally:
        shutil.rmtree(tempdir)


def parse_sourcedeps_line(line):
    line = line.strip()
    if not line or line.startswith("#"):
        return None, None, None
    local_path, branch = re.split("\s+", line, 1)
    branch = re.split("\s+", branch, 1)[0]
    revision = None
    if ";revno=" in branch:
        branch, revision = branch.rsplit(";revno=", 1)
    return local_path, branch, revision


def get_sourcedeps_content(branch, revision, path):
    text = branch.get_file_text(revision, path)
    if text is None:
        print("No file at %s" % path)
        return {}
    content = {}
    for line in text.splitlines():
        local_path, branch, revision = parse_sourcedeps_line(line)
        if local_path is not None:
            content[local_path] = (branch, revision)
    return content


def fixup_branch_location(branch_location):
    """Adjust a branch location to remove special hosts."""
    parsed_url = urlsplit(branch_location.encode('ascii'))
    if guess_vcs(branch_location) == VCS.VCS_BAZAAR:
        if (parsed_url.netloc.endswith('.launchpad.net') or
                # Handle a couple of special cases that have been used in
                # the past.
                parsed_url.netloc in ('bazaar.ubunet', 'bazaar.isd')):
            return "lp:" + parsed_url.path.lstrip("/")
        return branch_location
    else:
        components = list(parsed_url)
        if parsed_url.scheme == 'git+ssh' and not parsed_url.username:
            # Work around pygit2 being unable to supply username credentials.
            components[1] = '%s@%s' % (
                    settings.GIT_SSH_DEFAULT_USERNAME, parsed_url.hostname)
            port = parsed_url.port
            if port is None:
                port = settings.GIT_SSH_DEFAULT_PORT
            if port is not None:
                components[1] += ':%d' % port
        elif parsed_url.scheme == 'git+https':
            # codetree wants git+https://, but libgit2 wants https://.
            components[0] = 'https'
        return urlunsplit(components)


def sourcedeps_diff(old, new, branch_id, parent_branch_url,
        parent_branch_revision, fetched_branches=None):
    diff = []
    print("Old: %s\nNew: %s" % (old, new))
    for local_path, branch_info in new.items():
        print("Considering %s" % local_path)
        old_revspec = None
        tip_only = False
        new_branch_url, new_revspec = branch_info
        print("current sourcedeps line: branch=%s revno=%s" % (new_branch_url, new_revspec))
        if local_path in old:
            print("Present in old")
            old_branch_url, old_revspec = old[local_path]
            print("previous sourcedeps line: branch=%s revno=%s" % (old_branch_url, old_revspec))
            if (fixup_branch_location(old_branch_url) !=
                    fixup_branch_location(new_branch_url)):
                print("Changed branch from old, so doing one rev")
                tip_only = True
            elif old_revspec is None and new_revspec is None:
                if (fixup_branch_location(parent_branch_url) ==
                        fixup_branch_location(new_branch_url)):
                    print("Self-referring sourcedep, doing one rev")
                    tip_only = True
                    new_revspec = parent_branch_revision.revid
                else:
                    print("No idea about revisions")
                    continue
            elif old_revspec is None or new_revspec is None:
                print("Revno added or removed, so doing one rev")
                tip_only = True
        else:
            print("Not in old")
            tip_only = True
        new_branch_id = branch_id + local_path.replace("/", "_")
        branch = get_branch(new_branch_url, new_branch_id, allow_missing=True,
                fetched_branches=fetched_branches, all_revs=True)
        if branch is None:
            print("Branch couldn't be found, skipping")
            continue
        try:
            try:
                old_revision = branch.revspec_to_revision(old_revspec)
            except (bzr_errors.InvalidRevisionSpec, KeyError, ValueError):
                # Can't determine old revision, so just take one
                # revision
                old_revision = None
                tip_only = True
            new_revision = branch.revspec_to_revision(new_revspec)
            if (not tip_only and
                    old_revision is not None and new_revision is not None):
                if old_revision == new_revision:
                    print("Same in old and new, so skipping")
                    continue
                merged_rev = branch.find_lefthand_merger(
                        old_revision.revid, new_revision.revid)
                if merged_rev is None:
                    print("Went back to older revision, so doing one rev")
                    tip_only = True
            revisions = get_revisions(branch, new_branch_id, parent_branch_url,
                    start_revision=old_revision, end_revision=new_revision,
                    tip_only=tip_only)
            diff.extend([(branch, revision) for revision, _ in revisions])
        finally:
            print('Releasing lock for {}'.format(branch))
            branch.unlock()
    return diff


def sourcedeps_diff_revisions(branch, current_revision, previous_revision, sourcedeps_path,
        branch_id, parent_branch_url, fetched_branches=None):
    current_sourcedeps_content = get_sourcedeps_content(
            branch, current_revision, sourcedeps_path)
    previous_sourcedeps_content = get_sourcedeps_content(
            branch, previous_revision, sourcedeps_path)
    return sourcedeps_diff(previous_sourcedeps_content,
            current_sourcedeps_content, branch_id, parent_branch_url,
            current_revision, fetched_branches=fetched_branches)


def guess_vcs(branch_url):
    """Guess the VCS for a branch.

    We could set this explicitly in the database for top-level branches, but
    for source dependencies we have to guess anyway.  We therefore might as
    well always guess and do as good a job as possible.  codetree makes its
    decisions based on the URL scheme, so we do likewise.
    """
    parsed_url = urlsplit(branch_url.encode('ascii'))
    if parsed_url.scheme.startswith('git'):
        return VCS.VCS_GIT
    # Make the test suite's life easier.
    elif (parsed_url.scheme.endswith('file') and
            os.path.exists(os.path.join(unquote(parsed_url.path), '.git'))):
        return VCS.VCS_GIT
    else:
        return VCS.VCS_BAZAAR


def get_branch(branch_url, branch_id, git_branch=None, allow_missing=False,
        fetched_branches=None, all_revs=False):
    print("Getting branch %s" % branch_url)
    vcs = guess_vcs(branch_url)
    fixed_branch_url = fixup_branch_location(branch_url)
    try:
        branch = LocalBranch.get_or_create(vcs, fixed_branch_url, branch_id,
                git_branch=git_branch, fetched_branches=fetched_branches,
                all_revs=all_revs)
    except (bzr_errors.NotBranchError, bzr_errors.InvalidURL,
            pygit2.GitError) as e:
        if not allow_missing:
            raise
        print("Branch not found: %s" % e)
        return None
    print("Got branch")
    return branch


def get_revisions(branch, branch_id, parent_branch_url,
        start_revision=None, end_revision=None,
        sourcedeps_path=None, tip_only=False, fetched_branches=None):
    revisions = []
    for current_revision in branch.iter_revisions(
            start_revision=start_revision, end_revision=end_revision,
            tip_only=tip_only):
        sourcedep_changes = None
        previous_revision = branch.get_previous_revision(current_revision)
        if sourcedeps_path and previous_revision:
            print("Looking in %s for sourcedeps" % sourcedeps_path)
            sourcedep_changes = sourcedeps_diff_revisions(branch, current_revision,
                    previous_revision, sourcedeps_path, branch_id,
                    parent_branch_url, fetched_branches=fetched_branches)
            print("Changes in sourcedeps: %s" % sourcedep_changes)
        print("Adding rev %s" % current_revision.revno)
        revisions.append((current_revision, sourcedep_changes))
    return revisions


def update_revisions(project_id):
    load_plugins()
    project = Project.objects.get(pk=project_id)
    url = project.url
    last_revno = project.revision_set.filter(parent=None).aggregate(Max('revno'))['revno__max']
    fetched_branches = {}
    try:
        branch = get_branch(url, project.name, git_branch=project.git_branch,
            fetched_branches=fetched_branches)
    except bzr_errors.LockContention as err:
        print('Can not acquire lock for {}. There is already a lock at {}. '
              'Skipping.'.format(url, err.msg))
        return

    if last_revno is None:
        last_revision = None
    elif branch.vcs == VCS.VCS_BAZAAR:
        # In the Bazaar case, we don't store the revision ID, but we can ask
        # the branch to resolve the revno.
        last_revision = branch.resolve_revision(None, last_revno)
    else:
        # In the Git case, asking the branch to resolve the revno would be
        # an expensive operation, but we store the revision ID.
        last_revid = project.revision_set.filter(
                parent=None).get(revno=last_revno).commit_sha1
        last_revision = branch.get_revision(last_revid, revno=last_revno)
    try:
        revisions = get_revisions(branch, project.name, url,
                start_revision=last_revision,
                sourcedeps_path=project.sourcedeps_path,
                fetched_branches=fetched_branches)
        store_revisions(project, branch, revisions)
    finally:
        print('Releasing lock for {}'.format(branch))
        branch.unlock()


@shared_task
def update_revisions_task(project_id):
    update_revisions(project_id)


@shared_task
def update_all_projects():
    for project in Project.objects.all():
        # Have the subtask expire before the next one will be started to
        # prevent them piling up if the workers pause.
        expires = settings.CELERYBEAT_SCHEDULE['update_all_projects']['schedule'].seconds
        update_revisions_task.subtask((project.pk,), expires=expires).apply_async()
