import codecs
import json
import os

import bleach
from django.conf import settings
from django.utils.html import linebreaks
from django.utils.text import normalize_newlines
import markdown
import pybars
from pybars._compiler import escape
import waffle

from . import csrf


_compiler = None
def get_compiler():
    global _compiler
    if _compiler is None:
        _compiler = pybars.Compiler()
    return _compiler


_template_loader_cache = dict()
def load_template(name):
    global _template_loader_cache
    if name not in _template_loader_cache or settings.DEBUG:
        template_path = os.path.join(os.path.dirname(__file__),
                'templates', name)
        with codecs.open(template_path, 'r', 'utf-8') as f:
            content = f.read()
        _template_loader_cache[name] = content
    return _template_loader_cache[name]


_template_compiler_cache = dict()
def compile_template(name):
    global _template_compiler_cache
    if name not in _template_compiler_cache or settings.DEBUG:
        source = load_template(name)
        compiler = get_compiler()
        template = compiler.compile(source)
        _template_compiler_cache[name] = template
    return _template_compiler_cache[name]


def compile_template_from_string(name, source):
    global _template_compiler_cache
    if name not in _template_compiler_cache or settings.DEBUG:
        compiler = get_compiler()
        template = compiler.compile(source)
        _template_compiler_cache[name] = template
    return _template_compiler_cache[name]


def render_template(name, info, partials=None, helpers=None):
    template = compile_template(name)
    return template(info, partials=partials, helpers=helpers)


def list_templates():
    return [
        dict(name='deployment_request', content=load_template('revtracker/deployment_request.html')),
        dict(name='index', content=load_template('revtracker/index.html')),
        dict(name='project', content=load_template('revtracker/project.html')),
        dict(name='project_info', content=load_template('revtracker/project_info.html')),
        dict(name='revision_comment_form', content=load_template('revtracker/revision_comment_form.html')),
        dict(name='revision_info', content=load_template('revtracker/revision_info.html')),
        dict(name='revision_form', content=load_template('revtracker/revision_form.html')),
        dict(name='deployability_form', content=load_template('revtracker/deployability_form.html')),
        dict(name='editor', content=load_template('revtracker/editor.html')),
        ]


_partials_cache = None
def get_partials():
    global _partials_cache
    if _partials_cache is None or settings.DEBUG:
        templates = list_templates()
        partials = dict()
        for template in templates:
            partials[template['name']] = compile_template_from_string(
                    template['name'], template['content'])
        _partials_cache = partials
    return _partials_cache


def dump_json_helper(this, thing):
    while not getattr(thing, 'keys', None):
        thing = thing.context
    if 'global' in thing:
        thing = thing.copy()
        del thing['global']
    return json.dumps(thing)


def ignore_content(this, options):
    return choose_block(False, this, options)


def include_content(this, options):
    return choose_block(True, this, options)


def choose_block(condition, this, options):
    if condition:
        return options['fn'](this)
    else:
        return options['inverse'](this)


def csrf_token_helper(this, options):
    return csrf.get_csrf_token(options['helpers']['request'])


def current_url_helper(this, options):
    return options['helpers']['request'].path


def if_user_helper(this, options, condition):
    request = options['helpers']['request']
    attr = getattr(request.user, 'is_{}'.format(condition))
    return choose_block(attr, this, options)


def flag_template_helper(this, options, flag_name):
    request = options['helpers']['request']
    return choose_block(waffle.flag_is_active(request, flag_name),
            this, options)


def flag_info_helper(this, options):
    request = options['helpers']['request']
    flag_info = dict()
    # Shame that this performs pretty poorly with lots of flags
    for flag in waffle.models.Flag.objects.all():
        flag_info[flag.name] = waffle.flag_is_active(request, flag.name)
    return escape(json.dumps(flag_info))


def user_info_helper(this, options):
    user = options['helpers']['request'].user
    return escape(json.dumps({
        'authenticated': user.is_authenticated(),
        'active': user.is_active,
        'staff': user.is_staff,
        'superuser': user.is_superuser,
        }))


def linebreaks_helper(this, value):
    value = normalize_newlines(value).rstrip('\n')
    result = pybars.strlist()
    result.grow(linebreaks(value, autoescape=True))
    return result


def get_helpers(request):
   return {
        'csrf_token': csrf_token_helper,
        'current_url': current_url_helper,
        'dump_json': dump_json_helper,
        'flag': flag_template_helper,
        'flag_info': flag_info_helper,
        'if_user': if_user_helper,
        'linebreaks': linebreaks_helper,
        'request': request,
        'user_info': user_info_helper,
        }


def markdownize(text):
    # Expand tags allowed by bleach with some that our usual markdown
    # syntax generates
    allowed_tags = bleach.ALLOWED_TAGS
    allowed_tags.extend(('h1', 'h2', 'h3', 'p'))
    md_text = markdown.markdown(text, extensions=['urlize'])
    clean_text = bleach.clean(md_text, tags=allowed_tags)
    return clean_text


def markdownize_for_editor(info, key):
    editor_info = {}
    if info[key]:
        editor_info['raw'] = info[key]
        editor_info['formatted'] = markdownize(info[key])
    else:
        editor_info['raw'] = ''
        editor_info['formatted'] = "No description set"
    info[key] = editor_info
    return info
