import datetime

from django.core.urlresolvers import reverse
from django.contrib.auth.models import User
from django.contrib.comments.models import Comment
from django.contrib.contenttypes.models import ContentType
from django_factory import TestCase

from ..models import (
    DeploymentRequest,
    Project,
    Revision,
    )
from .. import data


class ProjectBasicInfoTests(TestCase):

    def test_includes_name(self):
        name = "project1"
        project = self.factory.make_one(Project, name=name)
        self.assertEqual(name, data.project_basic_info(project)['name'])

    def test_includes_description(self):
        description = "a description"
        project = self.factory.make_one(Project, description=description)
        self.assertEqual(description,
            data.project_basic_info(project)['description'])

    def test_null_description_becomes_blank(self):
        description = None
        project = self.factory.make_one(Project, description=description)
        self.assertEqual('',
            data.project_basic_info(project)['description'])

    def test_includes_url(self):
        name = "project1"
        project = self.factory.make_one(Project, name=name)
        self.assertEqual(reverse('project_view', args=(name,)),
            data.project_basic_info(project)['url'])

    def test_number_of_queries(self):
        project = self.factory.make_one(Project)
        from_db = Project.objects.get(pk=project.pk)
        self.assertNumQueries(0, data.project_basic_info, from_db)


class ProjectDeployedRevnoTests(TestCase):

    def test_None_when_not_deployed(self):
        project = self.factory.make_one(Project)
        self.assertEqual(None, data.project_deployed_revno(project))

    def make_deployed_project(self, revno=10):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project, revno=revno)
        project.deployed_revision = revision
        project.save()
        return project

    def test_with_deployed_revision(self):
        revno = 10
        project = self.make_deployed_project(revno)
        self.assertEqual(revno, data.project_deployed_revno(project))

    def test_number_of_queries(self):
        project = self.factory.make_one(Project)
        from_db = Project.objects.get(pk=project.pk)
        self.assertNumQueries(0, data.project_deployed_revno, from_db)

    def test_number_of_queries_deployed(self):
        project = self.make_deployed_project()
        from_db = Project.objects.get(pk=project.pk)
        self.assertNumQueries(1, data.project_deployed_revno, from_db)


class ProjectUndeployedRevisionsTests(TestCase):

    def test_no_revisions(self):
        project = self.factory.make_one(Project)
        deployed_revno = None
        self.assertEqual([],
                data.project_undeployed_revisions(project, deployed_revno))

    def test_revisions_only_for_other_projects(self):
        project = self.factory.make_one(Project)
        deployed_revno = None
        other_project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=other_project)
        self.assertEqual([],
                data.project_undeployed_revisions(project, deployed_revno))

    def test_no_deployed_revision(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        deployed_revno = None
        self.assertEqual([revision],
                data.project_undeployed_revisions(project, deployed_revno))

    def test_deploy_revision_is_last(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        project.deployed_revision = revision
        deployed_revno = revision.revno
        project.save()
        self.assertEqual([],
                data.project_undeployed_revisions(project, deployed_revno))

    def test_with_undeployed(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project, revno=1)
        revision2 = self.factory.make_one(Revision, project=project, revno=2)
        project.deployed_revision = revision1
        deployed_revno = revision1.revno
        project.save()
        self.assertEqual([revision2],
                data.project_undeployed_revisions(project, deployed_revno))

    def test_sorted(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project, revno=1)
        revision2 = self.factory.make_one(Revision, project=project, revno=2)
        revision3 = self.factory.make_one(Revision, project=project, revno=3)
        project.deployed_revision = revision1
        deployed_revno = revision1.revno
        project.save()
        self.assertEqual([revision2, revision3],
                data.project_undeployed_revisions(project, deployed_revno))

    def test_excludes_children(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project, revno=1)
        revision2 = self.factory.make_one(Revision, project=project,
                parent=revision1, revno=2)
        deployed_revno = None
        project.save()
        self.assertEqual([revision1],
            data.project_undeployed_revisions(project, deployed_revno))

    def test_number_of_queries(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project, revno=1)
        revision2 = self.factory.make_one(Revision, project=project, revno=2)
        revision3 = self.factory.make_one(Revision, project=project, revno=3)
        project.deployed_revision = revision1
        deployed_revno = revision1.revno
        project.save()
        from_db = Project.objects.get(pk=project.pk)
        self.assertNumQueries(1,
                data.project_undeployed_revisions, from_db, deployed_revno)


class ProjectDeployableRevisionsTests(TestCase):

    def test_project_deployable_revisions_no_revisions(self):
        self.assertEqual([], data.project_deployable_revisions([], {}))

    def test_project_deployable_revisions_undeployable_revision(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD)
        self.assertEqual([],
                data.project_deployable_revisions([revision], {}))

    def test_project_deployable_revisions_deployable_revision(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertEqual([revision],
                data.project_deployable_revisions([revision], {}))

    def test_project_deployable_revisions_only_those_after_deployable(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=1)
        revision2 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD, revno=2)
        revision3 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=3)
        undeployed_revisions = [revision1, revision2, revision3]
        self.assertEqual([revision1],
                data.project_deployable_revisions(undeployed_revisions, {}))

    def test_project_deployable_revisions_with_rollback(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD, revno=1)
        revision2 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=2)
        revision1.rolled_back_by = revision2
        revision1.save()
        undeployed_revisions = [revision1, revision2]
        self.assertEqual([revision1, revision2],
                data.project_deployable_revisions(undeployed_revisions, {}))

    def test_project_deployable_revisions_with_bad_child(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=1)
        child = self.factory.make_one(Revision, project=project,
                parent=revision1, status=Revision.STATUS_BAD, revno=1,
                branch="lp:foo")
        revision2 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=2)
        undeployed_revisions = [revision1, revision2]
        child_map = {revision1.pk: [child]}
        self.assertEqual([],
                data.project_deployable_revisions(undeployed_revisions,
                    child_map))

    def test_project_deployable_revisions_with_child_rollback(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=1)
        child = self.factory.make_one(Revision, project=project,
                parent=revision1, status=Revision.STATUS_BAD, revno=1,
                branch="lp:foo")
        revision2 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, revno=2)
        child.rolled_back_by = revision2
        child.save()
        undeployed_revisions = [revision1, revision2]
        child_map = {revision1.pk: [child]}
        self.assertEqual([revision1, revision2],
                data.project_deployable_revisions(undeployed_revisions,
                    child_map))

    def test_project_deployable_revisions_bad_rollback(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD, revno=1)
        revision2 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD, revno=2)
        revision1.rolled_back_by = revision2
        revision1.save()
        undeployed_revisions = [revision1, revision2]
        self.assertEqual([],
                data.project_deployable_revisions(undeployed_revisions, {}))

    def test_number_queries(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertNumQueries(0, data.project_deployable_revisions,
                [revision], {})

    def test_number_queries_with_rollback(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD)
        revision2 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        revision1.rolled_back_by = revision2
        revision1.save()
        self.assertNumQueries(0, data.project_deployable_revisions,
                [revision1, revision2], {})

    def test_number_queries_with_child(self):
        project = self.factory.make_one(Project)
        revision1 = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        child = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK, parent=revision1, branch="lp:foo")
        self.assertNumQueries(0, data.project_deployable_revisions,
                [revision1], {revision1.pk: [child]})


class MaximumDeployableRevisionTests(TestCase):

    def test_no_deployable(self):
        self.assertEqual(None, data.project_maximum_deployable_revision([]))

    def test_with_deployable(self):
        revno = 10
        revision = self.factory.make_one(Revision, revno=revno,
                status=Revision.STATUS_OK)
        self.assertEqual(revno,
                data.project_maximum_deployable_revision([revision]).revno)

    def test_last_deployable(self):
        revno = 11
        deployable = []
        deployable.append(self.factory.make_one(Revision, revno=revno-1,
                status=Revision.STATUS_OK))
        deployable.append(self.factory.make_one(Revision, revno=revno,
                status=Revision.STATUS_OK))
        self.assertEqual(revno,
                data.project_maximum_deployable_revision(deployable).revno)

    def test_number_queries(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_OK)
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(0,
                data.project_maximum_deployable_revision, [from_db])


class RevisionBlockingDeploymentTests(TestCase):

    def test_no_undeployed(self):
        self.assertEqual(None,
                data.project_revision_blocking_deployment([], []))

    def test_no_blockers(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_OK)
        self.assertEqual(None,
                data.project_revision_blocking_deployment([revision],
                    [revision]))

    def test_one_blocker(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_BAD)
        self.assertEqual(revision,
                data.project_revision_blocking_deployment([revision], []))

    def test_earliest_blocker(self):
        revision1 = self.factory.make_one(Revision, status=Revision.STATUS_BAD)
        revision2 = self.factory.make_one(Revision, revno=revision1.revno+1,
                status=Revision.STATUS_BAD)
        self.assertEqual(revision1,
                data.project_revision_blocking_deployment(
                    [revision1, revision2], []))

    def test_number_queries(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_BAD)
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(0,
                data.project_revision_blocking_deployment, [revision], [])


class ProjectRevisionsNeedingQATests(TestCase):

    def test_no_undeployed(self):
        self.assertEqual([], data.project_revisions_needing_qa([]))

    def test_undeployed_needing_qa(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN)
        self.assertEqual([revision],
                data.project_revisions_needing_qa([revision]))

    def test_undeployed_not_needing_qa(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_BAD)
        self.assertEqual([],
                data.project_revisions_needing_qa([revision]))

    def test_child_needing_qa(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_BAD)
        child = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN,  parent=revision,
                branch="lp:foo")
        self.assertEqual([child],
                data.project_revisions_needing_qa([revision]))

    def test_child_not_needing_qa(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_BAD)
        child = self.factory.make_one(Revision,
                status=Revision.STATUS_OK,  parent=revision,
                branch="lp:foo")
        self.assertEqual([],
                data.project_revisions_needing_qa([revision]))

    def test_parent_and_child_need_qa(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN)
        child = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN,  parent=revision,
                branch="lp:foo")
        self.assertEqual([child, revision],
                data.project_revisions_needing_qa([revision]))

    def test_number_queries(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN)
        self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN,  parent=revision,
                branch="lp:foo")
        self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN,  parent=revision,
                branch="lp:foo")
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(1,
                data.project_revisions_needing_qa, [from_db])


class RevisionBasicInfoTests(TestCase):

    def test_includes_pk(self):
        revision = self.factory.make_one(Revision)
        self.assertEqual(revision.pk, data.revision_basic_info(revision)['pk'])

    def test_includes_title(self):
        revno = 10
        revision = self.factory.make_one(Revision, revno=revno)
        self.assertEqual("Revision 10",
                data.revision_basic_info(revision)['title'])

    def test_includes_description(self):
        revno = 10
        revision = self.factory.make_one(Revision, revno=revno)
        self.assertEqual("revision 10",
                data.revision_basic_info(revision)['description'])

    def test_includes_branch(self):
        branch = "lp:foo"
        revision = self.factory.make_one(Revision, branch=branch)
        self.assertEqual(branch, data.revision_basic_info(revision)['branch'])

    def test_includes_branch_if_None(self):
        revision = self.factory.make_one(Revision, branch=None)
        self.assertEqual(None, data.revision_basic_info(revision)['branch'])

    def test_includes_revno(self):
        revno = 10
        revision = self.factory.make_one(Revision, revno=revno)
        self.assertEqual(revno, data.revision_basic_info(revision)['revno'])

    def test_includes_commit_sha1(self):
        commit_sha1 = "0" * 40
        revision = self.factory.make_one(Revision, commit_sha1=commit_sha1)
        self.assertEqual(commit_sha1,
                data.revision_basic_info(revision)['commit_sha1'])

    def test_includes_commit_sha1_if_None(self):
        revision = self.factory.make_one(Revision, commit_sha1=None)
        self.assertEqual(None,
                data.revision_basic_info(revision)['commit_sha1'])

    def test_includes_assignee(self):
        assignee = "james-w"
        revision = self.factory.make_one(Revision, assignee=assignee)
        self.assertEqual(assignee,
                data.revision_basic_info(revision)['assignee'])

    def test_includes_assignee_if_None(self):
        assignee = None
        revision = self.factory.make_one(Revision, assignee=assignee)
        self.assertEqual(assignee,
                data.revision_basic_info(revision)['assignee'])

    def test_includes_assignee_link(self):
        assignee = "james-w"
        revision = self.factory.make_one(Revision, assignee=assignee)
        self.assertEqual("https://launchpad.net/~{}".format(assignee),
                data.revision_basic_info(revision)['assignee_link'])

    def test_includes_assignee_link_if_None(self):
        assignee = None
        revision = self.factory.make_one(Revision, assignee=assignee)
        self.assertEqual(None,
                data.revision_basic_info(revision)['assignee_link'])

    def test_includes_link(self):
        link = "https://example.com/foo"
        revision = self.factory.make_one(Revision, link=link)
        self.assertEqual(link,
                data.revision_basic_info(revision)['link'])

    def test_includes_commit_message(self):
        commit_message = "This is a commit"
        revision = self.factory.make_one(Revision, commit_message=commit_message)
        self.assertEqual(commit_message,
                data.revision_basic_info(revision)['commit_message'])

    def test_number_queries(self):
        revision = self.factory.make_one(Revision)
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(0, data.revision_basic_info, from_db)


class RevisionSummaryStatusTests(TestCase):

    def test_revision_bad(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_BAD)
        child = self.factory.make_one(Revision, status=Revision.STATUS_UNKNOWN,
                parent=revision, branch="lp:foo")
        self.assertEqual(Revision.STATUS_BAD,
                data.revision_summary_status(revision, [child]))

    def test_revision_unknown(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN)
        child = self.factory.make_one(Revision, status=Revision.STATUS_BAD,
                parent=revision, branch="lp:foo")
        self.assertEqual(Revision.STATUS_UNKNOWN,
                data.revision_summary_status(revision, [child]))

    def test_revision_ok_no_chilren(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_OK)
        self.assertEqual(Revision.STATUS_OK,
                data.revision_summary_status(revision, []))

    def test_revision_ok_bad_child(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_OK)
        child = self.factory.make_one(Revision, status=Revision.STATUS_BAD,
                parent=revision, branch="lp:foo")
        self.assertEqual(Revision.STATUS_BAD,
                data.revision_summary_status(revision, [child]))

    def test_revision_ok_unknown_child(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_OK)
        child = self.factory.make_one(Revision, status=Revision.STATUS_UNKNOWN,
                parent=revision, branch="lp:foo")
        self.assertEqual(Revision.STATUS_UNKNOWN,
                data.revision_summary_status(revision, [child]))

    def test_revision_ok_untestable_child(self):
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_OK)
        child = self.factory.make_one(Revision, status=Revision.STATUS_UNTESTABLE,
                parent=revision, branch="lp:foo")
        self.assertEqual(Revision.STATUS_OK,
                data.revision_summary_status(revision, [child]))

    def test_number_queries(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_OK)
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(0, data.revision_summary_status, from_db, [])


class RevisionIsDeployableTests(TestCase):

    def test_unknown_revision_not_deployable(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_UNKNOWN)
        self.assertEqual(False, data.revision_is_deployable(revision, []))

    def test_bad_revision_not_deployable(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_BAD)
        self.assertEqual(False, data.revision_is_deployable(revision, []))

    def test_ok_revision_deployable(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_OK)
        self.assertEqual(True, data.revision_is_deployable(revision, []))

    def test_untestable_revision_deployable(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_UNTESTABLE)
        self.assertEqual(True, data.revision_is_deployable(revision, []))

    def test_child_revision_not_deployable(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_UNTESTABLE)
        child = self.factory.make_one(Revision, parent=revision,
                status=Revision.STATUS_BAD)
        self.assertEqual(False, data.revision_is_deployable(revision, [child]))

    def test_child_revision_deployable(self):
        revision = self.factory.make_one(Revision, status=Revision.STATUS_UNTESTABLE)
        child = self.factory.make_one(Revision, parent=revision,
                status=Revision.STATUS_OK)
        self.assertEqual(True, data.revision_is_deployable(revision, [child]))

    def test_number_queries(self):
        revision = self.factory.make_one(Revision)
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(0, data.revision_is_deployable, from_db, [])


class RevisionRolledBackByTests(TestCase):

    def test_not_rolled_back(self):
        revision = self.factory.make_one(Revision)
        self.assertEqual([], data.revision_rolled_back_by(revision, []))

    def test_rolled_back(self):
        revision = self.factory.make_one(Revision)
        rollback = self.factory.make_one(Revision)
        revision.rolled_back_by = rollback
        revision.save()
        self.assertEqual([rollback],
                data.revision_rolled_back_by(revision, []))

    def test_child_not_rolled_back(self):
        revision = self.factory.make_one(Revision)
        child = self.factory.make_one(Revision)
        self.assertEqual([], data.revision_rolled_back_by(revision, [child]))

    def test_child_rolled_back(self):
        revision = self.factory.make_one(Revision)
        child = self.factory.make_one(Revision)
        rollback = self.factory.make_one(Revision)
        child.rolled_back_by = rollback
        child.save()
        self.assertEqual([rollback],
                data.revision_rolled_back_by(revision, [child]))

    def test_parent_and_child_rolled_back(self):
        revision = self.factory.make_one(Revision)
        child = self.factory.make_one(Revision)
        parent_rollback = self.factory.make_one(Revision)
        child_rollback = self.factory.make_one(Revision)
        revision.rolled_back_by = parent_rollback
        revision.save()
        child.rolled_back_by = child_rollback
        child.save()
        self.assertEqual([parent_rollback, child_rollback],
                data.revision_rolled_back_by(revision, [child]))

    def test_number_queries(self):
        revision = self.factory.make_one(Revision)
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(0,
                data.revision_rolled_back_by, from_db, [])

    def test_number_queries_with_rollback(self):
        revision = self.factory.make_one(Revision)
        rollback = self.factory.make_one(Revision)
        revision.rolled_back_by = rollback
        revision.save()
        from_db = Revision.objects.get(pk=revision.pk)
        self.assertNumQueries(1,
                data.revision_rolled_back_by, from_db, [])


class GetChildRevisionsTests(TestCase):

    def test_no_parents(self):
        self.assertEqual({}, data.get_child_revisions([]))

    def test_no_children(self):
        revision = self.factory.make_one(Revision)
        self.assertEqual({}, data.get_child_revisions([revision]))

    def test_one_child(self):
        revision = self.factory.make_one(Revision)
        child = self.factory.make_one(Revision, parent=revision,
                branch="lp:foo")
        self.assertEqual({revision.pk: [child]},
                data.get_child_revisions([revision]))

    def test_two_children(self):
        revision = self.factory.make_one(Revision)
        child1 = self.factory.make_one(Revision, parent=revision,
                branch="lp:foo")
        child2 = self.factory.make_one(Revision, parent=revision,
                branch="lp:foo")
        self.assertEqual({revision.pk: [child1, child2]},
                data.get_child_revisions([revision]))

    def test_two_revisions_with_children(self):
        revision1 = self.factory.make_one(Revision)
        child1 = self.factory.make_one(Revision, parent=revision1,
                branch="lp:foo")
        revision2 = self.factory.make_one(Revision)
        child2 = self.factory.make_one(Revision, parent=revision2,
                branch="lp:foo")
        self.assertEqual({revision1.pk: [child1], revision2.pk: [child2]},
                data.get_child_revisions([revision1, revision2]))

    def test_number_queries(self):
        revision1 = self.factory.make_one(Revision)
        child1 = self.factory.make_one(Revision, parent=revision1,
                branch="lp:foo")
        revision2 = self.factory.make_one(Revision)
        child2 = self.factory.make_one(Revision, parent=revision2,
                branch="lp:foo")
        self.assertNumQueries(1,
                data.get_child_revisions, [revision1, revision2])


class CommentsForRevisionsTests(TestCase):

    def get_content_type(self, model):
        return ContentType.objects.get_for_model(model)

    def make_comment_for(self, model):
        return self.factory.make_one(Comment,
                content_type=self.get_content_type(model),
                object_pk=model.pk)

    def test_no_revisions(self):
        self.assertEqual({}, data.comments_for_revisions([]))

    def test_no_comments(self):
        revision = self.factory.make_one(Revision)
        self.assertEqual({}, data.comments_for_revisions([revision]))

    def test_one_comment(self):
        revision = self.factory.make_one(Revision)
        comment = self.make_comment_for(revision)
        self.assertEqual({revision.pk: [comment]},
                data.comments_for_revisions([revision]))

    def test_two_comments(self):
        revision = self.factory.make_one(Revision)
        comment1 = self.make_comment_for(revision)
        comment2 = self.make_comment_for(revision)
        self.assertEqual({revision.pk: [comment1, comment2]},
                data.comments_for_revisions([revision]))

    def test_two_revisions_with_comments(self):
        revision1 = self.factory.make_one(Revision)
        revision2 = self.factory.make_one(Revision)
        comment1 = self.make_comment_for(revision1)
        comment2 = self.make_comment_for(revision2)
        self.assertEqual({revision1.pk: [comment1], revision2.pk: [comment2]},
                data.comments_for_revisions([revision1, revision2]))

    def test_number_queries(self):
        revision1 = self.factory.make_one(Revision)
        revision2 = self.factory.make_one(Revision)
        comment1 = self.make_comment_for(revision1)
        comment2 = self.make_comment_for(revision2)
        self.assertNumQueries(1,
                data.comments_for_revisions, [revision1, revision2])


class GetParentRevnosTests(TestCase):

    def test_no_revisions(self):
        self.assertEqual({}, data.get_parent_revnos([]))

    def test_parent_revision(self):
        revision = self.factory.make_one(Revision)
        self.assertEqual({}, data.get_parent_revnos([revision]))

    def test_child_revision(self):
        revno = 30
        revision = self.factory.make_one(Revision, revno=revno)
        child = self.factory.make_one(Revision, parent=revision,
                branch="lp:foo", revno=1)
        self.assertEqual({child.pk: revno}, data.get_parent_revnos([child]))

    def test_multiple_reviions(self):
        revno = 30
        revision = self.factory.make_one(Revision, revno=revno)
        children = self.factory.make(2, Revision, parent=revision,
                branch="lp:foo", revno=1)
        expected = {
            children[0].pk: revno,
            children[1].pk: revno,
            }
        self.assertEqual(expected, data.get_parent_revnos(children))

    def test_query_count(self):
        revision = self.factory.make_one(Revision)
        children = self.factory.make(2, Revision, parent=revision,
                branch="lp:foo", revno=1)
        def load_rev(rev):
            return Revision.objects.get(pk=rev.pk)
        from_db = map(load_rev, children)
        self.assertNumQueries(1, data.get_parent_revnos, from_db)


class DeployedRevnosOfProjectsForRevisionsTests(TestCase):

    def test_no_revisions(self):
        self.assertEqual({},
                data.deployed_revnos_of_projects_for_revisions([]))

    def test_deployed(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        revno = 20
        deployed = self.factory.make_one(Revision, project=project,
                revno=revno)
        project.deployed_revision = deployed
        project.save()
        self.assertEqual({project.pk: revno},
                data.deployed_revnos_of_projects_for_revisions([revision]))

    def test_not_deployed(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        self.assertEqual({},
                data.deployed_revnos_of_projects_for_revisions([revision]))

    def test_query_count(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        deployed = self.factory.make_one(Revision, project=project)
        project.deployed_revision = deployed
        project.save()
        from_db = Revision.objects.select_related('project').get(pk=revision.pk)
        self.assertNumQueries(1,
                data.deployed_revnos_of_projects_for_revisions, [from_db])


class DeploymentRequestsTests(TestCase):

    def test_no_requests(self):
        self.assertEqual([], data.deployment_requests())

    def test_ignores_deployed(self):
        self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_COMPLETE)
        self.assertEqual([], data.deployment_requests())

    def test_includes_undeployed(self):
        request = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_WAITING)
        self.assertEqual(1, len(data.deployment_requests()))

    def test_data(self):
        request = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_WAITING)
        info = data.deployment_requests()[0]
        self.assertEqual(request.project.name, info['project'])
        self.assertEqual(request.date_created.strftime("%H:%M %Y-%m-%d"),
            info['date_created'])
        self.assertEqual(request.deploy_revision.revno,
            info['deploy_revision']['revno'])
        self.assertEqual(request.requester.username,
            info['requester'])
        self.assertEqual(
            reverse('deployment_request',
                    args=[request.project.name, request.pk]),
            info['link'])
        self.assertNotIn('completer', info)

    def test_order_by_date_requested(self):
        request2 = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_WAITING,
            date_created=datetime.datetime(2000, 1, 2))
        request1 = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_WAITING,
            date_created=datetime.datetime(2000, 1, 1))
        requests = data.deployment_requests()
        self.assertEqual(
            [request1.project.name, request2.project.name],
            [r['project'] for r in requests])

    def test_status_is_completed(self):
        completer = self.factory.make_one(User)
        request = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_COMPLETE,
            completer=completer, date_completed=datetime.datetime.now())
        info = data.deployment_requests(
            status=DeploymentRequest.STATUS_COMPLETE)[0]
        self.assertEqual(request.project.name, info['project'])
        self.assertEqual(request.date_created.strftime("%H:%M %Y-%m-%d"),
            info['date_created'])
        self.assertEqual(request.deploy_revision.revno,
            info['deploy_revision']['revno'])
        self.assertEqual(request.requester.username,
            info['requester'])
        self.assertEqual(request.date_completed.strftime("%H:%M %Y-%m-%d"),
            info['date_completed'])
        self.assertEqual(request.completer.username,
            info['completer'])
        self.assertEqual(
            reverse('deployment_request',
                    args=[request.project.name, request.pk]),
            info['link'])

    def test_order_by_date_completed(self):
        completer = self.factory.make_one(User)
        request1 = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_COMPLETE,
            date_completed=datetime.datetime(2000, 1, 1),
            completer=completer)
        request2 = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_COMPLETE,
            date_completed=datetime.datetime(2000, 1, 2),
            completer=completer)
        requests = data.deployment_requests(
            status=DeploymentRequest.STATUS_COMPLETE)
        self.assertEqual(
            [request2.project.name, request1.project.name],
            [r['project'] for r in requests])

    def test_query_count(self):
        expected_queries = 1
        self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_WAITING)
        self.assertNumQueries(expected_queries,
                data.deployment_requests)
        # Make a second request and check the number of queries doens't
        # change
        self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_WAITING)
        self.assertNumQueries(expected_queries,
                data.deployment_requests)

    def test_nice_string_representation(self):
        deploy_req = self.factory.make_one(DeploymentRequest,
            status=DeploymentRequest.STATUS_COMPLETE)
        expect_string = "DeploymentRequest - %s" % deploy_req.deploy_revision
        self.assertEqual(expect_string, str(deploy_req))
