from operator import itemgetter

from django.contrib.auth.models import User
from django.contrib.comments.models import Comment
from django.contrib.contenttypes.models import ContentType
from django_factory import TestCase
from testtools.matchers import Annotate

from .. import views
from ..models import Project, Revision


class TestIndexData(TestCase):

    def test_zero_projects(self):
        self.assertEqual(0, len(views.index_data()['projects']))

    def test_one_project(self):
        self.factory.make_one(Project)
        self.assertEqual(1, len(views.index_data()['projects']))

    def test_two_projects(self):
        self.factory.make(2, Project)
        self.assertEqual(2, len(views.index_data()['projects']))

    def test_includes_basic_info(self):
        name = "aproject"
        self.factory.make_one(Project, name=name)
        self.assertEqual(name, views.index_data()['projects'][0]['name'])

    def test_includes_undeployed_count(self):
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project)
        self.assertEqual(1,
                views.index_data()['projects'][0]['undeployed_count'])

    def test_includes_deployable_count(self):
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertEqual(1,
                views.index_data()['projects'][0]['deployable_count'])

    def test_includes_needing_qa_count(self):
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_UNKNOWN)
        self.assertEqual(1,
                views.index_data()['projects'][0]['needing_qa_count'])

    def test_query_count(self):
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertNumQueries(4, views.index_data)

    def test_query_count_multiple(self):
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        # FIXME: O(projects)
        self.assertNumQueries(7, views.index_data)


class GetProjectInfoTests(TestCase):

    def test_includes_project_info(self):
        project = self.factory.make_one(Project)
        self.assertIn('project', views.get_project_info(project))

    def test_includes_basic_info(self):
        name = "aproject"
        project = self.factory.make_one(Project, name=name)
        self.assertEqual(name,
                views.get_project_info(project)['project']['name'])

    def test_includes_deployed_revision(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        project.deployed_revision = revision
        project.save()
        self.assertEqual(revision.revno,
                views.get_project_info(project)['project']['deployed_revision']['revno'])

    def test_includes_maximum_deployable_revision(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertEqual(revision.revno,
                views.get_project_info(project)['project']['maximum_deployable_revision']['revno'])

    def test_includes_revision_blocking_deployment_if_exists(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_BAD)
        self.assertEqual(revision.revno,
                views.get_project_info(project)['project']['revision_blocking_deployment']['revno'])

    def test_excludes_revision_blocking_deployment_if_not_exists(self):
        project = self.factory.make_one(Project)
        self.assertNotIn('revision_blocking_deployment',
                views.get_project_info(project)['project'])

    def test_includes_undeployed_count(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        self.assertEqual(1,
                views.get_project_info(project)['project']['undeployed_count'])

    def test_includes_undeployed_revisions(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        self.assertEqual(1,
                len(views.get_project_info(project)['undeployed_revisions']))

    def test_revision_includes_basic_info(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        rev_info = views.get_project_info(project)['undeployed_revisions'][0]
        self.assertEqual(revision.pk, rev_info['pk'])

    def test_includes_child_revisions_when_none(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        rev_info = views.get_project_info(project)['undeployed_revisions'][0]
        self.assertEqual([], rev_info['child_revisions'])

    def test_includes_child_revisions_when_one(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        child = self.factory.make_one(Revision, parent=revision,
                branch="lp:foo")
        rev_info = views.get_project_info(project)['undeployed_revisions'][0]
        self.assertEqual(1, len(rev_info['child_revisions']))

    def test_includes_comments_for_child(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        child = self.factory.make_one(Revision, parent=revision,
                branch="lp:foo")
        self.factory.make_one(Comment, object_pk=child.pk,
                content_type=ContentType.objects.get_for_model(child))
        rev_info = views.get_project_info(project)['undeployed_revisions'][0]
        self.assertEqual(1, len(rev_info['child_revisions'][0]['comments']))

    def test_query_count(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertNumQueries(4, views.get_project_info, project)

    def test_query_count_two_revisions(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        self.assertNumQueries(4, views.get_project_info, project)

    def assertNumQueries(self, *args, **kwargs):
        from django.db import connection
        so_far = len(connection.queries)
        try:
            super(GetProjectInfoTests, self).assertNumQueries(*args, **kwargs)
        except AssertionError, e:
            e.mismatch.annotation += ", Queries were:\n%s" % "\n".join(
                map(itemgetter('sql'), connection.queries[so_far:]))
            raise e



class GetPersonInfoTests(TestCase):

    def test_includes_revisions_needing_qa(self):
        person = self.factory.make_one(User)
        self.assertIn('revisions_needing_qa', views.get_person_info(person))

    def test_no_revisions(self):
        person = self.factory.make_one(User)
        self.assertEqual(0,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_no_revisions_for_person(self):
        person = self.factory.make_one(User)
        other_person = self.factory.make_one(User)
        revision = self.factory.make_one(Revision, status=Revision.STATUS_UNKNOWN,
                committer=other_person)
        self.assertEqual(0,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_includes_unknown(self):
        person = self.factory.make_one(User)
        revison = self.factory.make_one(Revision, status=Revision.STATUS_UNKNOWN,
                committer=person)
        self.assertEqual(1,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_includes_bad(self):
        person = self.factory.make_one(User)
        revison = self.factory.make_one(Revision, status=Revision.STATUS_BAD,
                committer=person)
        self.assertEqual(1,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_excludes_ok(self):
        person = self.factory.make_one(User)
        revison = self.factory.make_one(Revision, status=Revision.STATUS_OK,
                committer=person)
        self.assertEqual(0,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_excludes_deployed(self):
        person = self.factory.make_one(User)
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN, committer=person)
        revision.project.deployed_revision = revision
        revision.project.save()
        self.assertEqual(0,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_excludes_before_deployed(self):
        person = self.factory.make_one(User)
        prev_revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN, committer=person, revno=1)
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN, committer=person, revno=2,
                project=prev_revision.project)
        revision.project.deployed_revision = revision
        revision.project.save()
        self.assertEqual(0,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_includes_undeployed(self):
        person = self.factory.make_one(User)
        deployed_revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN, committer=person, revno=1)
        deployed_revision.project.deployed_revision = deployed_revision
        deployed_revision.project.save()
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN, committer=person, revno=2)
        self.assertEqual(1,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def test_excludes_deployed_parent(self):
        person = self.factory.make_one(User)
        revision = self.factory.make_one(Revision,
                status=Revision.STATUS_OK, committer=person, revno=1)
        child = self.factory.make_one(Revision,
                status=Revision.STATUS_UNKNOWN, committer=person, revno=20,
                parent=revision, project=revision.project)
        revision.project.deployed_revision = revision
        revision.project.save()
        self.assertEqual(0,
                len(views.get_person_info(person)['revisions_needing_qa']))

    def get_rev_info(self, **kwargs):
        person = self.factory.make_one(User)
        kwargs.setdefault('status', Revision.STATUS_BAD)
        kwargs.setdefault('committer', person)
        revision = self.factory.make_one(Revision, **kwargs)
        return views.get_person_info(person)['revisions_needing_qa'][0]

    def test_revision_has_basic_info(self):
        commit_message = "the commit message"
        rev_info = self.get_rev_info(commit_message=commit_message)
        self.assertEqual(commit_message, rev_info['commit_message'])

    def test_revision_has_summary_status_display(self):
        rev_info = self.get_rev_info(status=Revision.STATUS_BAD)
        self.assertEqual(Revision.get_summary_status_display(Revision.STATUS_BAD),
                rev_info['summary_status_display'])

    def test_revision_has_project(self):
        project = self.factory.make_one(Project)
        rev_info = self.get_rev_info(project=project)
        self.assertEqual(project.name, rev_info['project']['name'])

    def test_revision_has_committer(self):
        # TODO: test the result better
        rev_info = self.get_rev_info()
        self.assertIn('committer', rev_info)

    def test_query_count(self):
        person = self.factory.make_one(User)
        self.factory.make_one(Revision, status=Revision.STATUS_UNKNOWN,
                committer=person)
        from_db = User.objects.get(pk=person.pk)
        self.assertNumQueries(5, views.get_person_info, person)

    def test_query_count_multiple_revisions(self):
        person = self.factory.make_one(User)
        self.factory.make(2, Revision, status=Revision.STATUS_UNKNOWN,
                committer=person)
        from_db = User.objects.get(pk=person.pk)
        self.assertNumQueries(5, views.get_person_info, person)
