import re
import soupmatchers

from django.core.urlresolvers import reverse


from ..models import Project, Revision

from . import(
    IndexProject,
    IndexProjectsList,
    RevTrackerTestCase,
    ProjectName,
    ProjectRevision,
    ProjectUpdateDeployedRevisionForm,
    ProjectDeployableRevision,
    ProjectDeployedRevision,
    ProjectBlockingRevision,
    ProjectRevisionChangeStatusForm,
    ProjectRevisionSetRollbackForm
    )

from test_api_helpers import User


class IndexViewTests(RevTrackerTestCase):

    urls = 'revtracker.urls'

    def setUp(self):
        super(IndexViewTests, self).setUp()
        self.create_user_and_log_in()

    def test_no_projects_has_no_projects(self):
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=IndexProject(count=0)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Tag('no projects message', 'div',
                  text="No projects registered.")))

    def test_one_project_has_one_project(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProjectsList(count=1), IndexProject(count=1))))

    def test_project_includes_name(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(), ProjectName(text=project.name))))

    def test_project_name_is_linked(self):
        name = "a-name"
        project = self.factory.make_one(Project, name=name)
        url = reverse('project_view', args=(project.name,))
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(
                  soupmatchers.Tag("project link", "a", attrs=dict(href=url)),
                  ProjectName(text=name))))

    def test_number_of_undeployed_revisions_is_included(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(),
                  soupmatchers.Tag('undeployed revisions', 'div', attrs={'class':'undeployed three columns'},
                      text=re.compile("1 undeployed revisions")))))

    def test_number_of_undeployed_revisions_is_included_when_zero(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(),
                  soupmatchers.Tag('undeployed revisions', 'div', attrs={'class':'undeployed three columns'},
                      text=re.compile("0 undeployed revisions")))))

    def test_number_of_deployable_revisions_is_included(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_OK)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(),
                  soupmatchers.Tag('deployable revisions', 'div', attrs={'class':'deployable three columns'},
                      text=re.compile("1 deployable revision")))))

    def test_number_of_deployable_revisions_is_included_when_zero(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(),
                  soupmatchers.Tag('deployable revisions', 'div', attrs={'class':'deployable three columns'},
                      text=re.compile("0 deployable revisions")))))

    def test_number_of_revisions_needing_qa_is_included(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project,
                status=Revision.STATUS_UNKNOWN)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(),
                  soupmatchers.Tag('revisions needing QA', 'div', attrs={'class':'qa_needed three columns omega'},
                      text=re.compile("1 revisions need QA")))))

    def test_number_of_revisions_needing_qa_is_included_when_zero(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('index'))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Within(IndexProject(),
                  soupmatchers.Tag('revisions needing QA', 'div', attrs={'class':'qa_needed three columns omega'},
                      text=re.compile("0 revisions need QA")))))


class ProjectViewTests(RevTrackerTestCase):

    urls = 'revtracker.urls'

    def setUp(self):
        super(ProjectViewTests, self).setUp()
        self.create_user_and_log_in()

    def test_404_for_non_existant_project(self):
        response = self.client.get(reverse('project_view', args=('missing',)))
        self.assertEqual(404, response.status_code)

    def test_project_view_shows_project_name(self):
        name = "a-name"
        project = self.factory.make_one(Project, name=name)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Tag('project name', 'h1', text=name)))

    def test_project_view_shows_deployed_revision_unknown(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectDeployedRevision(text="Currently deployed: unknown")))

    def test_project_view_shows_deployed_revision_known(self):
        project = self.factory.make_one(Project)
        revno = 100
        revision = self.factory.make_one(Revision, revno=revno, project=project)
        project.deployed_revision = revision
        project.save()
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectDeployedRevision(text="Currently deployed: %s" % revision.description)))

    def test_project_view_all_revisions_deployed(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectDeployableRevision(text="All revisions are deployed",
                attrs={'class': "global_info status_good"})))

    def test_project_view_no_deployable_revisions(self):
        project = self.factory.make_one(Project)
        self.factory.make_one(Revision, project=project, status=Revision.STATUS_BAD)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectDeployableRevision(text="No revisions are currently ready to be deployed")))

    def test_project_view_deployable_revisions(self):
        project = self.factory.make_one(Project)
        revno = 100
        revision = self.factory.make_one(Revision, project=project,
                revno=revno, status=Revision.STATUS_OK)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectDeployableRevision(text="Revision %d is ready to be deployed" % revno)))

    def test_shows_no_blocking_revision(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectBlockingRevision(count=0)))

    def test_shows_no_blocking_revision_if_there_is_one(self):
        project = self.factory.make_one(Project)
        revno = 100
        revision = self.factory.make_one(Revision, project=project,
                revno=revno, status=Revision.STATUS_BAD)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectBlockingRevision(text="Revision %d is blocking deployment" % revno)))

    def test_project_one_undeployed_revision(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(html_matches=ProjectRevision(revision)))

    def test_includes_link(self):
        project = self.factory.make_one(Project)
        revno = 100
        link = 'http://example.com/alink'
        revision = self.factory.make_one(Revision, project=project, link=link, revno=revno)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag('revision link', 'a', attrs={'href': revision.link},
                text="Revision %d" % revno).within(ProjectRevision(revision))))

    def test_shows_commit_message(self):
        project = self.factory.make_one(Project)
        commit_message = self.factory.getUniqueUnicode()
        revision = self.factory.make_one(Revision, project=project, commit_message=commit_message)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag('commit message', 'div', attrs={'class': 'commit_message fifteen columns alpha'},
                text=commit_message).within(ProjectRevision(revision))))

    def test_shows_rollback(self):
        project = self.factory.make_one(Project)
        rollback_revno = 11
        revision = self.factory.make_one(Revision, project=project, revno=rollback_revno-1)
        revision.rolled_back_by = self.factory.make_one(Revision, project=project, revno=rollback_revno)
        revision.save()
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag('rollback message', 'div', attrs={'class': 'rollback noaction_status fifteen columns alpha'},
                text="Rolled back by revision %d" % rollback_revno).within(ProjectRevision(revision))))

    def test_shows_rollback_candidates(self):
        """<option> for rollback candidates contains revision's title."""
        project = self.factory.make_one(Project)
        rollback_revno = 11
        # A bad revision, so that the list of candidates is rendered
        revision = self.factory.make_one(
            Revision, project=project, revno=rollback_revno-1,
            status=Revision.STATUS_BAD)
        # This one should appear in the list
        revision.rolled_back_by = self.factory.make_one(
            Revision, project=project, revno=rollback_revno)
        revision.save()
        response = self.client.get(
            reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag(
                'rollback candidate', 'option',
                attrs={'value': revision.rolled_back_by.pk},
                text=revision.rolled_back_by.title).within(
                    ProjectRevision(revision))))

    def test_shows_committer(self):
        project = self.factory.make_one(Project)
        name = 'bob'
        other_name = 'jane'
        committer = self.factory.make_one(User, username=name)
        revision = self.factory.make_one(Revision, project=project, committer=committer, assignee=other_name)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag('assignee link', 'a', attrs={'href': reverse('person_view', args=(name,))},
                text=name).within(soupmatchers.Tag('assignee', 'div', attrs={'class': 'assignee main six columns alpha'})).within(
                        ProjectRevision(revision))))

    def test_shows_nothing_if_assignee_and_committer_missing(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project, assignee=None, committer=None)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag('assignee', 'div', attrs={'class': 'assignee main six columns alpha'},
                text=re.compile("(unknown)")).within(
                    ProjectRevision(revision))))

    def test_shows_assignee_if_committer_is_none(self):
        project = self.factory.make_one(Project)
        name = self.factory.getUniqueUnicode()
        revision = self.factory.make_one(Revision, project=project, assignee=name, committer=None)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=soupmatchers.Tag('assignee link', 'a', attrs={'href': revision.assignee_link},
                text=name).within(soupmatchers.Tag('assignee', 'div', attrs={'class': 'assignee main six columns alpha'})).within(
                        ProjectRevision(revision))))

    def test_shows_revision_status_edit_form(self):
        project = self.factory.make_one(Project)
        revno = 100
        revision = self.factory.make_one(Revision, project=project, revno=revno)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectRevisionChangeStatusForm(revision).within(ProjectRevision(revision))))

    def test_shows_rollback_form_for_bad_revision(self):
        project = self.factory.make_one(Project)
        revno = 100
        revision = self.factory.make_one(Revision, project=project, revno=revno,
            status=Revision.STATUS_BAD)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectRevisionSetRollbackForm(revision).within(ProjectRevision(revision))))

    def test_project_shows_update_deployed_revision_form(self):
        project = self.factory.make_one(Project)
        response = self.client.get(reverse('project_view', args=(project.name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
            html_matches=ProjectUpdateDeployedRevisionForm(project)))


class ProjectUpdateRevisionTests(RevTrackerTestCase):

    urls = 'revtracker.urls'

    def setUp(self):
        super(ProjectUpdateRevisionTests, self).setUp()
        self.create_user_and_log_in()

    def test_404_for_non_existant_project(self):
        response = self.client.post(reverse('project_view', args=('missing',)))
        self.assertEqual(404, response.status_code)

    def DISABLED_test_errors_for_invalid_revno(self):
        # FIXME: needs porting to pybars style
        project = self.factory.make_one(Project)
        response = self.client.post(reverse('project_view', args=(project.name,)),
                data={"revno": "a"})
        self.assertEqual(200, response.status_code)
        self.assertFormError(response, 'update_deployed_revision_form', 'revno', 'Invalid revision number')

    def DISABLED_test_errors_for_non_existant_revision(self):
        # FIXME: needs porting to pybars style
        project = self.factory.make_one(Project)
        response = self.client.post(reverse('project_view', args=(project.name,)),
                data={"revno": 499})
        self.assertEqual(200, response.status_code)
        self.assertFormError(response, 'update_deployed_revision_form', 'revno', 'Unknown revision')

    def DISABLED_test_updates_deployed_revision(self):
        # FIXME: needs porting to pybars style
        project = self.factory.make_one(Project)
        revno = 10
        revision = self.factory.make_one(Revision, project=project, revno=revno)
        response = self.client.post(reverse('project_view', args=(project.name,)),
                data={"revno": revno})
        self.assertRedirects(response, reverse('project_view', args=(project.name,)))
        project = Project.objects.get(pk=project.pk)
        self.assertEqual(revision, project.deployed_revision)


class ProjectChangeRevisionStatusTests(RevTrackerTestCase):

    urls = 'revtracker.urls'

    def setUp(self):
        super(ProjectChangeRevisionStatusTests, self).setUp()
        self.create_user_and_log_in()

    def test_404_on_non_existant_project(self):
        response = self.client.post(reverse('project_change_revision_status', args=('missing', '44')))
        self.assertEqual(404, response.status_code)

    def test_404_on_non_existant_revision(self):
        project = self.factory.make_one(Project)
        response = self.client.post(reverse('project_change_revision_status', args=(project.name, '44')))
        self.assertEqual(404, response.status_code)

    def test_302_on_get(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        response = self.client.get(reverse('project_change_revision_status', args=(project.name, revision.pk)))
        self.assertRedirects(response, reverse('project_view', args=(project.name,)))

    def test_updates_status(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project, status=Revision.STATUS_UNKNOWN)
        response = self.client.post(reverse('project_change_revision_status', args=(project.name, revision.pk)),
                data=dict(status=Revision.STATUS_OK))
        revision = Revision.objects.get(pk=revision.pk)
        self.assertEqual(Revision.STATUS_OK, revision.status)

    def test_invalid_submission(self):
        project = self.factory.make_one(Project)
        revision = self.factory.make_one(Revision, project=project)
        response = self.client.post(reverse('project_change_revision_status', args=(project.name, revision.pk)),
                data=dict(status="nonsense"))
        self.assertEqual(400, response.status_code)


class PersonViewTests(RevTrackerTestCase):

    urls = 'revtracker.urls'

    def setUp(self):
        super(PersonViewTests, self).setUp()
        self.create_user_and_log_in()

    def test_404_for_non_existant_person(self):
        response = self.client.get(reverse('person_view', args=('missing',)))
        self.assertEqual(404, response.status_code)

    def test_person_view_shows_person_name(self):
        name = "a-name"
        project = self.factory.make_one(User, username=name)
        response = self.client.get(reverse('person_view', args=(name,)))
        self.assertThat(response, soupmatchers.HTMLResponseHas(
              html_matches=soupmatchers.Tag('person name', 'h1', text=name)))
